% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation, version 2.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Point lifting (unified projection model   %
%         for omnidirectional sensor )         %
%                                              %
%  Created : 03/09/2004                        %
%   Author : Christopher Mei                   %
% Based on : Barreto PhD 2003                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%% Input %%%%%%%%%%%%%%%%%%%%%%%
% Xi : points to lift
% p : latus rectum
% d : distance between focal points (any for paraboloid)
% f : scale factor of camera
% type : Paraboloid (1)/Hyperboloid (2)/Ellipsoid (3)

function [P,dPdX] = pointLifting(Xi, p, d, f, type)

if nargin==0
  disp('Launching test...');
  test;
  return
end

%Conic parameters
if type==1
  xi=1;
  phi=1+2*p;
end
if type==2
  xi=d/sqrt(d^2+4*p^2);
  phi=(d+2*p)/sqrt(d^2+4*p^2);
end
if type==3
  xi=d/sqrt(d^2+4*p^2);
  phi=(d-2*p)/sqrt(d^2+4*p^2);
end

switch nargout
 case 0
 case 1
  P=lift(Xi, xi, phi, f, type);
 case 2
  P=lift(Xi, xi, phi, f, type);
  dPdX = jacobianXFull(Xi, xi, phi);
 otherwise
  disp('Too many output args.')
end

function test
error=1e-4;
disp(['Error : 1e' slog10(error)]);

%%% Values %%%
X=10*randn(2,1);
dX=10*randn(2,1)*error;

p=16.7;
d=0; 
f = 1;
type=1;

[P,dPdX]=pointLifting(X,p,d,f,type);

%%% Test derivation in X %%%
Pp=pointLifting(X+dX,p,d,f,type);
dP=dPdX*dX;

N=norm(Pp-P-dP);

if N~=0
  gainP=norm(Pp-P)/N;
  disp(['Estimate gain in P : 1e' slog10(gainP)]);
else
  disp(['Estimate gain in P : Inf']);
end

function P=lift(Xi, xi, phi, f, type)
if size(Xi,1)==2
  Xi=[Xi(1,:);Xi(2,:);ones(1,size(Xi,2))];
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Calculate Bcircle

Mc=[[xi-phi,0,0];
    [0,xi-phi,0];
    [0,0,1]];

if type==1
  Bcircle = inv(Mc)*Xi;
end
if type==2 | type==3
  Rc=[[1,0,0];
      [0,1,0];
      [0,0,1]];
  Kc=[[f,0,0];
      [0,f,0];
      [0,0,1]];
  Hc=Kc*Rc*Mc;
  invHc=inv(Hc);
  Bcircle=invHc*Xi;
end

%Calculate the lifting of Bcircle on the sphere
P=invhslash(Bcircle,xi);

%%%%%%%%%%%%%%%%%% Calculate lifting %%%%%%%%%%%%%%%%%%%

function Bmirror=invhslash(M,xi)
nb=size(M);
nb=nb(2);
Bmirror=zeros(3,nb);

for i=1:nb
  x=M(1,i);
  y=M(2,i);
  z=M(3,i);
  lambda=(-z*xi-sqrt(z^2+(1-xi^2)*(x^2+y^2)))/(x^2+y^2+z^2);
  %signLambda=sign(lambda)
  Bmirror(1,i)=lambda*x;
  Bmirror(2,i)=lambda*y;
  Bmirror(3,i)=lambda*z+xi;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function dPdX = jacobianXFull(X,xi,phi)
[m,n] = size(X);
dPdX=zeros(3*n,m);

for i=1:n
  dPdX(3*(i-1)+1:3*i,1:m)=jacobianX(X(:,i),xi,phi);
end

function dPdX = jacobianX(X,xi,phi)

m=size(X,1);

if m==2
  X=[X(1,:);X(2,:);ones(1,size(X,2))];
end

invMc=inv([[xi-phi,0,0];
	   [0,xi-phi,0];
	   [0,0,1]]);

X = invMc*X;

x=X(1); y=X(2); z=X(3);

n2=x^2+y^2+z^2;
s=sqrt(xi^2*z^2+(1-xi^2)*n2);
t=s+xi*z;

lambda=-t/n2;
dlamdbadX=1/(s*n2^2)*[t^2*x t^2*y (t^2*z-xi*n2*t)];

dPdX=(X*dlamdbadX+lambda*eye(3))*invMc;

if m==2
  dPdX=dPdX(:,1:2);
end

function s=slog10(val)
s=num2str(floor(log10(val)));
