% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation, version 2.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%


% go_omni_calib_optim_iter
%
% Main calibration function. Computes the intrinsic and extrinsic parameters.
% Runs as a script.
%
%INPUT: x_1,x_2,x_3,...: Feature locations on the images
%       X_1,X_2,X_3,...: Corresponding grid coordinates
%
%OUTPUT: fc: Camera focal length
%        cc: Principal point coordinates
%        alpha_c: Skew coefficient
%        kc: Distortion coefficients
%        KK: The camera matrix (containing fc and cc)
%        omc_1,omc_2,omc_3,...: 3D rotation vectors attached to the grid positions in space
%        Tc_1,Tc_2,Tc_3,...: 3D translation vectors attached to the grid positions in space
%        Rc_1,Rc_2,Rc_3,...: 3D rotation matrices corresponding to the omc vectors
%
%Method: Minimizes the pixel reprojection error in the least squares sense over the intrinsic
%        camera parameters, and the extrinsic parameters (3D locations of the grids in space)
%
%Note: If the intrinsic camera parameters (fc, cc, kc) do not exist before, they are initialized through
%      the function init_intrinsic_param.m. Otherwise, the variables in memory are used as initial guesses.
%
%Note: The row vector active_images consists of zeros and ones. To deactivate an image, set the
%      corresponding entry in the active_images vector to zero.
%

if ~exist('desactivated_images'),
    desactivated_images = [];
end;

if ~exist('est_aspect_ratio'),
    est_aspect_ratio = 1;
end;

if ~exist('est_fc');
    est_fc = [1;1]; % Set to zero if you do not want to estimate the focal length (it may be useful! believe it or not!)
end;

if ~exist('recompute_extrinsic'),
    recompute_extrinsic = 1; % Set this variable to 0 in case you do not want to recompute the extrinsic parameters
    % at each iteration.
end;

if ~exist('check_cond'),
    check_cond = 1; % Set this variable to 0 in case you don't want to extract view dynamically
end;

if ~exist('center_optim'),
    center_optim = 1; %%% Set this variable to 0 if your do not want to estimate the principal point
end;

if exist('est_dist'),
    if length(est_dist) == 4,
        est_dist = [est_dist ; 0];
    end;
end;

if ~exist('est_dist'),
    est_dist = [1;1;1;1;0];
end;

if ~exist('est_alpha'),
    est_alpha = 0; % by default, do not estimate skew
end;


% Little fix in case of stupid values in the binary variables:
center_optim = double(~~center_optim);
est_alpha = double(~~est_alpha);
est_dist = double(~~est_dist);
est_fc = double(~~est_fc);
est_aspect_ratio = double(~~est_aspect_ratio);



fprintf(1,'\n');

if ~exist('nx')&~exist('ny'),
  fprintf(1,'WARNING: No image size (nx,ny) available. Setting nx=640 and ny=480. If these are not the right values, change values manually.\n');
  nx = 640;
  ny = 480;
end;


%check_active_images;

if ~exist('dont_ask'),
  dont_ask = 0;
end;


% A quick fix for solving conflict
if ~isequal(est_fc,[1;1]),
    est_aspect_ratio=1;
end;
if ~est_aspect_ratio,
    est_fc=[1;1];
end;


if ~est_aspect_ratio,
    fprintf(1,'Aspect ratio not optimized (est_aspect_ratio = 0) -> fc(1)=fc(2). Set est_aspect_ratio to 1 for estimating aspect ratio.\n');
else
    if isequal(est_fc,[1;1]),
        fprintf(1,'Aspect ratio optimized (est_aspect_ratio = 1) -> both components of fc are estimated (DEFAULT).\n');
    end;
end;

if ~isequal(est_fc,[1;1]),
    if isequal(est_fc,[1;0]),
        fprintf(1,'The first component of focal (fc(1)) is estimated, but not the second one (est_fc=[1;0])\n');
    else
        if isequal(est_fc,[0;1]),
            fprintf(1,'The second component of focal (fc(1)) is estimated, but not the first one (est_fc=[0;1])\n');
        else
            fprintf(1,'The focal vector fc is not optimized (est_fc=[0;0])\n');
        end;
    end;
end;


if ~center_optim, % In the case where the principal point is not estimated, keep it at the center of the image
    fprintf(1,'Principal point not optimized (center_optim=0). ');
    if ~exist('cc'),
        fprintf(1,'It is kept at the center of the image.\n');
        cc = [(nx-1)/2;(ny-1)/2];
    else
        fprintf(1,'Note: to set it in the middle of the image, clear variable cc, and run calibration again.\n');
    end;
else
    fprintf(1,'Principal point optimized (center_optim=1) - (DEFAULT). To reject principal point, set center_optim=0\n');
end;


if ~center_optim & (est_alpha),
    fprintf(1,'WARNING: Since there is no principal point estimation (center_optim=0), no skew estimation (est_alpha = 0)\n');
    est_alpha = 0;  
end;

if ~est_alpha,
    fprintf(1,'Skew not optimized (est_alpha=0) - (DEFAULT)\n');
    alpha_c = 0;
else
    fprintf(1,'Skew optimized (est_alpha=1). To disable skew estimation, set est_alpha=0.\n');
end;


if ~exist('alpha_c'),
    fprintf(1,'Initialization of the image skew to zero.\n');
    alpha_c = 0;
    alpha_smooth = 0.4; % slow convergence
end;

if ~exist('kc'),
    fprintf(1,'Initialization of the image distortion to zero.\n');
    kc = zeros(5,1);
    alpha_smooth = 0.4; % slow convergence
end;

if ~est_aspect_ratio,
  if ~exist('fc')
    if exist('KK_estimate')
      fc=[KK_estimate(1,1);KK_estimate(2,2)];
      cc=[KK_estimate(1:2,3)];
    else
      disp(['Please estimate the parameters using the mirror border' ...
	    ' first.']);
    end
  end
  fc(1) = (fc(1)+fc(2))/2;
  fc(2) = fc(1);
end;

if ~prod(double(est_dist)),
    % If no distortion estimated, set to zero the variables that are not estimated
    kc = kc .* est_dist;
end;


if ~prod(double(est_fc)),
    fprintf(1,'Warning: The focal length is not fully estimated (est_fc ~= [1;1])\n');
end;

% Put the initial estimates in param
if ~exist('KK_estimate')&~exist('fc')
  disp(['Please estimate the parameters using the mirror border' ...
	' first.']);
  return
else
  if exist('fc')
    XI=[kc;alpha_c;fc;cc];
  else
    % Initialise the distortions with 0 and the other values with
    % the estimation using the mirror border
    XI=[zeros(5,1);0;KK_estimate(1,1);KK_estimate(2,2);KK_estimate(1,3);KK_estimate(2,3)];
  end
end
%

%XI

%return
param=[XI;zeros(7*n_ima,1)];

for kk = 1:length(ind_active),     
  index = ind_active(kk);
  eval(['Qw_kk = Qw_' num2str(index) ';']);
  eval(['Tw_kk = Tw_' num2str(index) ';']);
  if isnan(Qw_kk(1)),
    fprintf(1,'Extrinsic parameters at frame %d do not exist\n',kk);
    return;
  end;
  param(10+7*(kk-1) + 1:10+7*(kk-1) + 7)=[Qw_kk;Tw_kk];
end

%-------------------- Main Optimization:

fprintf(1,['\nMain calibration optimization procedure - Number of' ...
	   ' images : %d\n'], length(ind_active));



fprintf(1,'Gradient descent iterations : ');

fc = XI(7:8);
cc = XI(9:10);

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Optimisation settings %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
iter = 0;
cond_thresh = 1e-30;

VERS = version;
VERS =  VERS(1);

if(VERS=='7')
  disp(['WARNING: removing singular matrix warning and managing it' ...
	' internally.'])
  warning('off','MATLAB:nearlySingularMatrix')
end
  
emax1=1e-10;

if ~exist('taux')
  taux=0.01;
end

if ~exist('nu')
  nu=2;
end

if ~exist('MaxIterBiased')
  MaxIterBiased = 100;
end
if ~exist('freqRecompExtrBiased')
  freqRecompExtrBiased=10;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The following vector helps to select the 
% variables to update (for only active images):
selected_variables = [est_dist; est_alpha;est_fc;center_optim*ones(2,1);...
		    reshape(ones(7,1)*active_images,7*n_ima,1)];
%selected_variables = [est_dist; est_alpha;est_fc;center_optim*ones(2,1);...
%		    reshape(zeros(7,1)*active_images,7*n_ima,1)];



if ~est_aspect_ratio,
  if isequal(est_fc,[1;1]) | isequal(est_fc,[1;0]),
    selected_variables(8) = 0;
  end;
end;
%selected_variables=ones(7*n_ima+10,1);
%selected_variables'
ind_Jac = find(selected_variables)';
%%%%%%%%%%%%%%%%%%%%%%%%%%%

buildJacobian;
mu=taux*max(max(JJ3));
found=(max(abs(ex3))<emax1);

%size(JJ3)
%return
do_recomp=1;

while ~found&(iter<MaxIterBiased),
  fprintf(1,'%d...',iter+1);
  
  JJ3=JJ3+mu*eye(size(JJ3,1));

  if rcond(JJ3)<cond_thresh
    disp('Matrix badly conditionned, stopping...')
    break
  end

  if ~est_aspect_ratio & isequal(est_fc,[1;1]),
    param(8) = param(7);
  end;

  %size(JJ3)
  JJ3_old = JJ3;
  ex3_old = ex3;
  
  JJ3 = JJ3(ind_Jac,ind_Jac);
  ex3 = ex3(ind_Jac);
  
  hlm=-inv(JJ3)*ex3;
  
  param_old = param;
  param(ind_Jac) = param(ind_Jac)+hlm;
  buildValue;
  sFx=norm(sfx)^2;
  sFxp1=norm(sfxp1)^2;
  
  quote=(sFx-sFxp1)/(0.5*hlm'*(mu*hlm-ex3));
  
  if quote>0
    buildJacobian;
    found=max(abs(ex3))<emax1;
    mu=mu*max(1/3,1-(2*quote-1)^3);
    nu=2; 
    do_recomp = 1;
  else
    JJ3 = JJ3_old;
    ex3 = ex3_old;
    param = param_old;
    mu=mu*nu;
    nu=2*nu;
  end
      
  %% Second step: (optional) - It makes convergence faster, and the region of convergence LARGER!!!
  %% Recompute the extrinsic parameters only using compute_extrinsic.m (this may be useful sometimes)
  %% The complete gradient descent method is useful to precisely update the intrinsic parameters.
  
  if recompute_extrinsic&(mod(iter+1,freqRecompExtrBiased)==0) %==0,  
    if do_recomp
      do_recomp=0;
      fprintf(1,'(r) ');

      for kk = 1:ind_active 
	%kk
	Qw_current = param(10+7*(kk-1) + 1:10+7*(kk-1) + 4);
	Tw_current = param(10+7*(kk-1) + 5:10+7*(kk-1) + 7);
	eval(['X_kk = X_' num2str(kk) ';']);
	eval(['x_kk = x_' num2str(kk) ';']);
	xp=omniCamProjection(-X_kk,[Qw_current;Tw_current;param(1:10)],...
			     mirror_p, mirror_d, mirror_type);
	
	error_init = mean(mean(abs(xp-x_kk),2));
	[Qw_new,Tw_new,error,k]=fastOmniPnPBiased(-X_kk, x_kk, [Qw_current;Tw_current;param(1:10)], ...
						  mirror_p, mirror_d, ...
						  mirror_type);
	%error
	xp=omniCamProjection(-X_kk,[Qw_new;Tw_new;param(1:10)],...
			     mirror_p, mirror_d, mirror_type);
	error_new = mean(mean(abs(xp-x_kk),2));
	if check_cond
	  if error_new/error_init>5
	    active_images(kk) = 0;
	    fprintf(1,'\nWarning: View #%d is causing problems. This image is now set inactive. (note: to disactivate this option, set check_cond=0)\n',kk);
	    desactivated_images = [desactivated_images kk];
	    Qw_new = NaN*ones(4,1);
	    Tw_new = NaN*ones(3,1); 
	    check_active_images;
	  end;
	end
	param(10+7*(kk-1) + 1:10+7*(kk-1) + 4) = Qw_new;
	param(10+7*(kk-1) + 5:10+7*(kk-1) + 7) = Tw_new;
      end
    end;
  end;

  iter = iter + 1;    
end;
JJ2_inv=inv(JJ3);

fprintf(1,'done\n');



%%%--------------------------- Computation of the error of estimation:

fprintf(1,'Estimation of uncertainties...');


%check_active_images;

solution = param;


% Extraction of the parameters for computing the right reprojection error:
kc = solution(1:5);
alpha_c = solution(6);
fc = solution(7:8);
cc = solution(9:10);

KK=[fc(1) fc(1)*alpha_c cc(1);
    0 fc(2) cc(2);
    0 0 1];

for kk = 1:length(ind_active)
  index = ind_active(kk);

  Qw_kk = solution(10+7*(kk-1) + 1: 10+7*(kk-1) + 4);
  Tw_kk = solution(10+7*(kk-1) + 5: 10+7*(kk-1) + 7);
      
  eval(['Qw_' num2str(index) ' = Qw_kk;']);
  eval(['Tw_' num2str(index) ' = Tw_kk;']);
end;


% Recompute the error (in the vector ex):
comp_omni_error;
comp_omni_sphere_error;

sigma_x = std(ex(:));

param_error =  3*sqrt(full(diag(JJ2_inv)))*sigma_x;

kc_error = param_error(1:5);
alpha_c_error = param_error(6);
fc_error = param_error(7:8);
cc_error = param_error(9:10);

fprintf(1,'done\n');

show_intrinsic;

%%% Some recommendations to the user to reject some of the difficult unkowns... Still in debug mode.

alpha_c_min = alpha_c - alpha_c_error/2;
alpha_c_max = alpha_c + alpha_c_error/2;

if (alpha_c_min < 0) & (alpha_c_max > 0),
    fprintf(1,'Recommendation: The skew coefficient alpha_c is found to be equal to zero (within its uncertainty).\n');
    fprintf(1,'                You may want to reject it from the optimization by setting est_alpha=0 and run Calibration\n\n');
end;

kc_min = kc - kc_error/2;
kc_max = kc + kc_error/2;

prob_kc = (kc_min < 0) & (kc_max > 0);

if ~(prob_kc(3) & prob_kc(4))
    prob_kc(3:4) = [0;0];
end;


if sum(prob_kc),
    fprintf(1,'Recommendation: Some distortion coefficients are found equal to zero (within their uncertainties).\n');
    fprintf(1,'                To reject them from the optimization set est_dist=[%d;%d;%d;%d;%d] and run Calibration\n\n',est_dist & ~prob_kc);
end;


return;


%function value=gain(hlm,mu,g)
%value=1/2*hlm'*(mu*hlm-g);
