% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation, version 2.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unified projection model for omnidirectional %
%          sensors and jacobian                %
%                                              %
%  Created : 2005                              %
%   Author : Christopher Mei                   %
% Based on : Barreto PhD 2003                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%% Input %%%%%%%%%%%%%%%%%%%%%%%
%X : points to project, 3xN matrix
%p : latus rectum
%d : distance between focal points (any for paraboloid)
%type : Perspective (0)/Paraboloid (1)/Hyperboloid (2)/Ellipsoid (3)

function [P,dPdX] = fastUnifiedPointProjection(X, p, d, type)

%path(path,'/home/cmei/prog_nosave/Matlab/General');

if nargin==0
  disp('Launching test...');
  test;
  return
end

[xi,xiMphi]=pd2xixiMphi(p,d,type);

switch nargout
 case 0
 case 1
  P=transform(X,xi,xiMphi);
 case 2
  P=transform(X,xi,xiMphi);
  dPdX = jacobianMirrorFull(X,xi,xiMphi);
 otherwise
  disp('Too many output args.')
end

function test
error=1e-2;

X=randn(3,1);
dX=randn(3,1)*error;
p=3;
d=20;
type=2;

disp(['Error : ' num2str(error)]);

[P,dPdX]=fastUnifiedPointProjection(X,p,d,type);
Pp=fastUnifiedPointProjection(X+dX,p,d,type);

n=size(X,2);
P=reshape(P,2*n,1);
Pp=reshape(Pp,2*n,1);
dP=dPdX*dX;

%Pp-P-dP
errorP=norm(Pp-P-dP);

disp(['Estimate error : ' num2str(errorP)]);

function P=transform(X,xi,xiMphi)
nb=size(X);
nb=nb(2);

P=zeros(2,nb);

for col=1:nb
  x=X(1,col);
  y=X(2,col);
  z=X(3,col);
  denom=xiMphi/(z-xi*sqrt(x^2+y^2+z^2));
  P(1,col)=x*denom;
  P(2,col)=y*denom;
end

function dPdX = jacobianMirrorFull(X,xi,xiMphi)
[m,n] = size(X);
dPdX=zeros(2*n,3);

for i=1:n
  dPdX(2*(i-1)+1:2*i,:)=jacobianMirror(X(:,i),xi,xiMphi);
end

function dPdX = jacobianMirror(X,xi,xiMphi)
x=X(1); y=X(2); z=X(3); r=norm(X);

dPdX=xiMphi/(r*(z-xi*r)^2)*...
     [r*z-xi*(y^2+z^2), xi*x*y, x*(xi*z-r);...
      xi*x*y, r*z-xi*(x^2+z^2), y*(xi*z-r)];

function [xi,xiMphi]=pd2xixiMphi(p,d,type)
%Conic parameters
switch type
 case 0
  xi = 0;
  xiMphi = -1;
 case 1
  xi=1;
  xiMphi=-2*p;
 case 2
  denom=sqrt(d^2+4*p^2);
  xi=d/denom;
  xiMphi=(-2*p)/denom;
 case 3
  denom=sqrt(d^2+4*p^2);
  xi=d/denom;
  xiMphi=(2*p)/denom;
 otherwise
  disp('Unknown conic type');
end



