% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation, version 2.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                 %
%     Calculates the camera projection function   %
%               and its jacobian                  %
%                                                 %
%   Created : 2005                                %
%    Author : Christopher Mei                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [P,dPdV] = omniCamProjection(X, V, p, d, type)

if nargin==0
  disp('Launching test...');
  test;
  return
end

switch nargout
 case 0
 case 1
  P=transform(X, V, p, d, type);
 case 2
  [P,dPdV]=transformFull(X, V, p, d, type);
 otherwise
  disp('Too many output args.')
end

function test
error=1e-4;
disp(['Error : 1e' slog10(error)]);

%%% Values %%%
X=randn(3,4);

V=randn(17,1);
dV=randn(17,1)*error;

p=1;
d=NaN;
type=1;

[P,dPdV]=omniCamProjection(X,V,p,d,type);

Pp=omniCamProjection(X,V+dV,p,d,type);

dP=dPdV*dV;

nb=size(X,2);
Pp=reshape(Pp,2*nb,1);
P=reshape(P,2*nb,1);

gainP=norm(Pp-P)/norm(Pp-P-dP);

disp(['Estimate gain in X : 1e' slog10(gainP)]);

function P=transform(X, V, p, d, type)

P=rigid_motion_quat(X, V(1:4), V(5:7));
P=fastUnifiedPointProjection(P, p, d, type);
P=distortion(P, V(8:12));
P=pointProjection(P, V(13:17));

function [P,dPdV]=transformFull(X, V, p, d, type)
[P,dWdQ,dWdT]=rigid_motion_quat(X, V(1:4), V(5:7));
dWdV1=[dWdQ dWdT];
[P,dHdW]=fastUnifiedPointProjection(P, p, d, type);
[P,dDdV2,dDdH]=distortion(P, V(8:12));
[P,dPdV3,dPdD]=pointProjection(P, V(13:17));

% $$$ sdPdD=size(dPdD)
% $$$ sdDdH=size(dDdH)
% $$$ sdHdW=size(dHdW)
% $$$ sdWdV1=size(dWdV1)
% $$$ sdDdV2=size(dDdV2)
% $$$ sdPdV3=size(dPdV3)

nb=size(X,2);
dPdV=zeros(2*nb,17);

for i=1:nb
  inter=2*(i-1)+1:2*i;
  dPdV(inter,:)=[dPdD(inter,:)*[dDdH(inter,:)*dHdW(inter,:)*dWdV1(3*(i-1)+1:3*i,:) dDdV2(inter,:)] dPdV3(inter,:)];
end

%dPdV=[dPdD*[dDdH*dHdW*dWdV1, dDdV2], dPdV3];

size(dPdV);

function dPdV = jacobianVProjectionFull(X,alpha,f1,f2,c1,c2)
[m,n] = size(X);
dPdV=zeros(2*n,5);

for i=1:n
  dPdV(2*(i-1)+1:2*i,:)=jacobianVProjection(X(:,i),alpha,f1,f2,c1,c2);
end

function dPdX = jacobianXProjectionFull(X,alpha,f1,f2,c1,c2)
[m,n] = size(X);
dPdX=zeros(2*n,2);

for i=1:n
  dPdX(2*(i-1)+1:2*i,:)=jacobianXProjection(X(:,i),alpha,f1,f2,c1,c2);
end

function dPdV = jacobianVProjection(X,alpha,f1,f2,c1,c2)
x=X(1); y=X(2);

dPdV=[f1*y,x+alpha*y,0,1,0;
      0,0,y,0,1];

function dPdX = jacobianXProjection(X,alpha,f1,f2,c1,c2)
dPdX=[f1,f1*alpha;
      0,f2];

function s=slog10(val)
s=num2str(floor(log10(val)));
