% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation, version 2.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software Foundation,
% Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Calls a C program to undistort images     %
%                                             %
%   Created : 2005                            %
%    Author : Christopher Mei                 %
%                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if ~exist('biased_calib')
  biased_calib=1;
end

if ~exist('omni')
  omni=0;
end

if ((biased_calib|~omni)&(~exist('fc')|~exist('cc')|~exist('kc')|...
			  ~exist('alpha_c')))|...
      (~biased_calib&(~exist('fc_sphere')|~exist('cc_sphere')|~exist('kc_sphere')|~exist('alpha_c_sphere')))
  
  fprintf(1,'No intrinsic camera parameters available.\n');
  return;
end
   
if omni&(~exist('center_estimate')|~exist('radius_estimate'))
  fprintf(1,'Please estimate first the radius and center of the mirror border.\n');
  return;
end

if biased_calib|~omni
  disp('Biased calib (ie normal).');
  KK = makeKK([alpha_c;fc;cc]);
else
  disp('Unbiased calib (VERY SLOW).');
  KK_sphere = makeKK([alpha_c_sphere;fc_sphere;cc_sphere]);
end

if ~omni
  %%% Compute the new KK matrix to fit as much data in the image (in order to
  %%% accomodate large distortions:
  r2_extreme = (nx^2/(4*fc(1)^2) + ny^2/(4*fc(2)^2));
  dist_amount = 1; %(1+kc(1)*r2_extreme + kc(2)*r2_extreme^2);
  fc_new = dist_amount * fc;

  KK_new = makeKK([alpha_c;fc_new;cc]);
else
  if biased_calib
    % Computer new focals and ROI to fit as much data as possible in the image
    [cf1,cf2,roi_min,roi_max]=calculateFocalsBiased(center_estimate,radius_estimate,kc,alpha_c,fc,cc);
  else
    [cf1_sphere,cf2_sphere,roi_min_sphere,roi_max_sphere]=calculateFocalsUnbiased(center_estimate,...
						  radius_estimate,kc_sphere,alpha_c_sphere,fc_sphere,cc_sphere);
  end
  
  % Ask the user if he wants the focals to be the same (this preserves
  % symmetry)
  same_focals = 1;
  
  same_focals = input('Should we impose the same focals ? ([]=yes, other=no) : ','s');
  
  if isempty(same_focals)
    same_focals = 1;
    fprintf(1,'Same focals.\n');
  else
    same_focals = 0;
    disp('Different focals.');
  end
  
  if same_focals
    cf1=max(cf1,cf2);
    cf2=cf1;
  end
  
  if biased_calib
    KK_new=makeKK([0;cf1;cf2;cc]);
  else
    invKK_sphere=inv(makeKK([alpha_c_sphere;fc_sphere;cc_sphere]));
    KK_new_sphere=makeKK([0;cf1_sphere;cf2_sphere;invKK_sphere(1:2,3)]);
  end
  
  [height,width] = size(I);
  
  create_luts = input('Should we create/recreate LUTs ? ([]=yes, other=no) : ','s');
  
  if isempty(create_luts)
    fprintf(1,'Creating LUTs... ');
    if biased_calib
      write_lut_biased(width,height,kc,alpha_c,fc,cc,KK_new);
      write_params_to_file('undistProjModel',KK_new, ...
			   mirror_p, mirror_d, mirror_type,...
			   roi_min, roi_max);
      write_params_to_file_ascii('undistProjModel.m',KK_new, ...
				 mirror_p, mirror_d, mirror_type,...
				 roi_min, roi_max);
    else
      write_lut_unbiased(width,height,kc_sphere,alpha_c_sphere,fc_sphere,cc_sphere,KK_new_sphere);
      write_params_to_file('undistProjModel_sphere',KK_new_sphere, ...
			   mirror_p, mirror_d, mirror_type,...
			   roi_min_sphere,roi_max_sphere);
      write_params_to_file_ascii('undistProjModel_sphere.m',KK_new_sphere, ...
				 mirror_p, mirror_d, mirror_type,...
				 roi_min_sphere,roi_max_sphere);
    end
  
    fprintf(1,'done\n');
  else
    disp('Using current LUTs (if they exist).');
  end
  
  
  bilinear = input('Bilinear [] or nearest neighbour (other) ? ','s');
  
  if isempty(bilinear)
    bilinear = 1;
    fprintf(1,'Bilinear.\n');
  else
    bilinear = 0;
    disp('Nearest neighbour.');
  end
end

quest = input('Do you want to undistort all the calibration images ([],0) or a new image (1)? ');

if isempty(quest),
    quest = 0;
end;

if ~quest,
    
    if n_ima == 0,
        fprintf(1,'No image data available\n');
        return;
    end;
    
    if ~exist(['I_' num2str(ind_active(1))]),
        ima_read_calib;
    end;
    
    check_active_images;   
    
    format_image2 = format_image;
    if format_image2(1) == 'j',
        format_image2 = 'bmp';
    end;
    
    for kk = 1:n_ima,
        
        if exist(['I_' num2str(kk)]),
            
            eval(['I = I_' num2str(kk) ';']);
	    
	    if ~omni
	      [I2] = rect2(I,eye(3),fc,cc,kc,KK_new);
	    end
	    
            if ~type_numbering,   
                number_ext =  num2str(image_numbers(kk));
            else
                number_ext = sprintf(['%.' num2str(N_slots) 'd'],image_numbers(kk));
            end;
            
	    
            ima_name2 = [calib_name '_rect' number_ext '.' format_image2];
            
            fprintf(1,['Saving undistorted image under ' ima_name2 '...\n']);
            
            if omni
	      firstName = [calib_name number_ext '.' format_image2];
	      
	      if bilinear
		if biased_calib
		  system([undistortPath ' -b1 lutx luty ' firstName ...
			  ' ' ima_name2]);
		  
		else
		  system([undistortPath ' -b1 lutx_sphere luty_sphere ' firstName ...
			  ' ' ima_name2]);
		end
	      else
		if biased_calib
		  system([undistortPath ' -n lutx_round luty_round ' ...
			  firstName ' ' ima_name2]);
		else
		  system([undistortPath ' -n lutx_round_sphere luty_round_sphere ' ...
			  firstName ' ' ima_name2]);
		  
		end
	      end
	      if 1
		disp(['Reprojecting points with new (simplified)' ...
		      ' calibration model...']);
		sindex = num2str(kk);
		eval(['x=x_' sindex ';']);
		eval(['X=X_' sindex ';']);
		eval(['Qw=Qw_' sindex ';']);
		eval(['Tw=Tw_' sindex ';']);
		
		I2=imread(ima_name2); 
		
		figure(kk+1);
		if exist('imshow')&exist('uint8')
		  imshow(uint8(I2));
		else
		  image(I2);
		  colormap(gray(256));
		end
		xtr=rigid_motion_quat(-X, Qw, Tw);
		xtr=fastUnifiedPointProjection(xtr, mirror_p,...
					       mirror_d, mirror_type);
		
		if biased_calib
		  x_new=KK_new*[xtr;ones(1,size(xtr,2))];
		else
		  x_new=KK_new_sphere*[xtr;ones(1,size(xtr,2))];
		end
		hold on;
		plot(x_new(1,:),x_new(2,:),'m+');
		drawnow;
	      end
	    else
	      if format_image2(1) == 'p',
                if format_image2(2) == 'p',
		  saveppm(ima_name2,uint8(round(I2)));
                else
		  savepgm(ima_name2,uint8(round(I2)));
                end;
	      else
                if format_image2(1) == 'r',
		  writeras(ima_name2,round(I2),gray(256));
                else
		  imwrite(uint8(round(I2)),gray(256),ima_name2,format_image2);
                end;
	      end;
	    end
                        
        end;
        
    end;
    
    fprintf(1,'done\n');
    
else
    
    dir;
    fprintf(1,'\n');
    
    image_name = input('Image name (full name without extension): ','s');
    
    format_image2 = '0';
    
    while format_image2 == '0',
        
        format_image2 =  input('Image format: ([]=''r''=''ras'', ''b''=''bmp'', ''t''=''tif'', ''p''=''pgm'', ''j''=''jpg'', ''m''=''ppm'') ','s');
        
        if isempty(format_image2),
            format_image2 = 'ras';
        end;
        
        if lower(format_image2(1)) == 'm',
            format_image2 = 'ppm';
        else
            if lower(format_image2(1)) == 'b',
                format_image2 = 'bmp';
            else
                if lower(format_image2(1)) == 't',
                    format_image2 = 'tif';
                else
                    if lower(format_image2(1)) == 'p',
                        format_image2 = 'pgm';
                    else
                        if lower(format_image2(1)) == 'j',
                            format_image2 = 'jpg';
                        else
                            if lower(format_image2(1)) == 'r',
                                format_image2 = 'ras';
                            else  
                                disp('Invalid image format');
                                format_image2 = '0'; % Ask for format once again
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;
    
    ima_name = [image_name '.' format_image2];

    %%% READ IN IMAGE:
    
    if format_image2(1) == 'p',
      if format_image2(2) == 'p',
	I = double(loadppm(ima_name));
      else
	I = double(loadpgm(ima_name));
      end;
    else
      if format_image2(1) == 'r',
	I = readras(ima_name);
      else
	I = double(imread(ima_name));
      end;
    end;
    
    if size(I,3)>1,
      I = I(:,:,2);
    end;
    
    if (size(I,1)>ny)|(size(I,2)>nx),
      I = I(1:ny,1:nx);
    end;

    %% SHOW THE ORIGINAL IMAGE:
    
    figure(2);
    if exist('imshow')&exist('uint8')
      imshow(uint8(I));
    else
      image(I);
      colormap(gray(256));
    end
    
    title('Original image (with distortion) - Stored in array I');
    drawnow;
        
    ima_name2 = [image_name '_rect.' format_image2];
    
    fprintf(1,'Computing the undistorted image...')
      
    if ~omni
      %% UNDISTORT THE IMAGE:
      [I2] = rect(I,eye(3),fc,cc,kc,alpha_c,KK_new);
    else
      if bilinear
	callSys = [undistortPath ' -b1 lutx luty ' ima_name ' ' ima_name2];
      else
	callSys = [undistortPath ' -n lutx_round luty_round ' ima_name ...
		   ' ' ima_name2];
      end
      system(callSys);
      I2 = imread(ima_name2);
    end      
    
    fprintf(1,'done\n');
        
    figure(3);
    if exist('imshow')&exist('uint8')
      imshow(uint8(I2));
    else
      image(I2);
      colormap(gray(256));
    end
    title('Undistorted image - Stored in array I2');
    drawnow;
    
    if ~omni
      %% SAVE THE IMAGE IN FILE:
    
      %ima_name2 = [image_name '_rect.' format_image2];
      
      fprintf(1,['Saving undistorted image under ' ima_name2 '...']);
      
      if format_image2(1) == 'p',
        if format_image2(2) == 'p',
	  saveppm(ima_name2,uint8(round(I2)));
        else
	  savepgm(ima_name2,uint8(round(I2)));
        end;
      else
        if format_image2(1) == 'r',
	  writeras(ima_name2,round(I2),gray(256));
        else
	  imwrite(uint8(round(I2)),gray(256),ima_name2,format_image2);
        end;
      end;
      
      fprintf(1,'done\n');
    end
end

